#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <sys/vt.h>
#include <sys/stat.h>
#include <errno.h>
#include <signal.h>
#include <sys/kd.h>
#include <asm/io.h>
#include "wisio.h"
#define GRAPH_BASE 0xA0000
#define GRAPH_SIZE 0x10000
int width,height;
unsigned char *GM;


int __svgalib_mem_fd = -1;	/* /dev/mem file descriptor  */
int __svgalib_tty_fd = -1;	/* /dev/tty file descriptor */
static int svgalib_vc = -1, startup_vc = -1;

static struct termios text_termio;	/* text mode termio parameters     */
static struct termios graph_termio;	/* graphics mode termio parameters */
struct vt_mode __svgalib_oldvtmode;
static int my_pid = 0;		/* process PID, used with atexit() */

setioperm()
{
        if (ioperm(0x3b4, 0x3df - 0x3b4 + 1, 1)) 
	    exit(1);
}        

static void restoretextmode(void)
{
    if (__svgalib_tty_fd >= 0) {
	ioctl(__svgalib_tty_fd, KDSETMODE, KD_TEXT);
        ioctl(__svgalib_tty_fd, VT_SETMODE, &__svgalib_oldvtmode);
    }
}



static void __vga_atexit(void)
{
    if (getpid() == my_pid)	/* protect against forked processes */
	restoretextmode();
    if (__svgalib_tty_fd >= 0 && startup_vc > 0)
	    ioctl(__svgalib_tty_fd, VT_ACTIVATE, startup_vc);
}



static void __vga_mmap(void)
{
    GM = (unsigned char *) mmap(
				   (caddr_t) 0,
				   GRAPH_SIZE,
				   PROT_READ | PROT_WRITE,
				   MAP_SHARED,
				   __svgalib_mem_fd,
				   GRAPH_BASE
	);
}

void __svgalib_takevtcontrol(void)
{
    struct sigaction siga;
    struct vt_mode newvtmode;

    ioctl(__svgalib_tty_fd, VT_GETMODE, &__svgalib_oldvtmode);
    newvtmode = __svgalib_oldvtmode;
    newvtmode.mode = VT_PROCESS;	/* handle VT changes */
    newvtmode.relsig = SIGUSR1;	/* I didn't find SIGUSR1/2 anywhere */
    newvtmode.acqsig = SIGUSR2;	/* in the kernel sources, so I guess */
    /* they are free */
/*    SETSIG(siga, SIGUSR1, __svgalib_releasevt_signal);
    SETSIG(siga, SIGUSR2, __svgalib_acquirevt_signal); */
    ioctl(__svgalib_tty_fd, VT_SETMODE, &newvtmode);
}

void __svgalib_waitvtactive(void)
{
    if (__svgalib_tty_fd < 0)
	return; /* Not yet initialized */

    while (ioctl(__svgalib_tty_fd, VT_WAITACTIVE, svgalib_vc) < 0) {
	if ((errno != EAGAIN) && (errno != EINTR)) {
 	    perror("ioctl(VT_WAITACTIVE)");
	    exit(1);
	}
	usleep(150000);
    }
}

/* open /dev/mem */
static void open_mem(void)
{
    if (__svgalib_mem_fd < 0)
	if ((__svgalib_mem_fd = open("/dev/mem", O_RDWR)) < 0) {
	    printf("svgalib: Cannot open /dev/mem.\n");
	    exit(1);
	}
}

void __svgalib_open_devconsole(void)
{
    struct vt_mode vtm;
    struct vt_stat vts;
    struct stat sbuf;
    char fname[30];

    if (__svgalib_tty_fd >= 0)
        return;

    /*
     * Now, it would be great if we could use /dev/tty and see what it is connected to.
     * Alas, we cannot find out reliably what VC /dev/tty is bound to. Thus we parse
     * stdin through stderr for a reliable VC
     */
    for (__svgalib_tty_fd = 0; __svgalib_tty_fd < 3; __svgalib_tty_fd++) {
        if (fstat(__svgalib_tty_fd, &sbuf) < 0)
            continue;
        if (ioctl(__svgalib_tty_fd, VT_GETMODE, &vtm) < 0)
            continue;
        if ((sbuf.st_rdev & 0xff00) != 0x400)
            continue;
        if (!(sbuf.st_rdev & 0xff))
            continue;
        svgalib_vc = sbuf.st_rdev & 0xff;
        return;                 /* perfect */
    }

    if ((__svgalib_tty_fd = open("/dev/console", O_RDWR)) < 0) {
        printf("svgalib: can't open /dev/console \n");
        exit(1);
    }
    if (ioctl(__svgalib_tty_fd, VT_OPENQRY, &svgalib_vc) < 0)
        goto error;
    if (svgalib_vc <= 0)
        goto error;
    sprintf(fname, "/dev/tty%d", svgalib_vc);
    close(__svgalib_tty_fd);
    /* change our control terminal: */
    setsid();
    /* We must use RDWR to allow for output... */
    if (((__svgalib_tty_fd = open(fname, O_RDWR)) >= 0) &&
        (ioctl(__svgalib_tty_fd, VT_GETSTATE, &vts) >= 0)) {
        fflush(stdin);
        fflush(stdout);
        fflush(stderr);
        close(0);
        close(1);
        close(2);
        dup(__svgalib_tty_fd);
        dup(__svgalib_tty_fd);
        dup(__svgalib_tty_fd);
        /* clear screen and switch to it */
        fwrite("\e[H\e[J", 6, 1, stderr);
        fflush(stderr);
        if (svgalib_vc != vts.v_active) {
            startup_vc = vts.v_active;
	    ioctl(__svgalib_tty_fd, VT_ACTIVATE, svgalib_vc);
            __svgalib_waitvtactive();
	}
    } else {
error:
    if (__svgalib_tty_fd > 2)
	close(__svgalib_tty_fd);
    __svgalib_tty_fd = - 1;
    printf("Not running in a graphics capable console,\n"
	 "and unable to find one.\n");
    }
}

static void vgad_exit(void);

static void initialize(void)
{
    int i;
    struct sigaction siga;

    __svgalib_open_devconsole();
    if (__svgalib_tty_fd < 0) {
	exit(1);
    }

    /* Make sure that textmode is restored at exit(). */
    if (my_pid == 0)
	my_pid = getpid();
    atexit(vgad_exit);

#ifndef DONT_WAIT_VC_ACTIVE
/*    __svgalib_waitvtactive();*/
#endif

    /* save text mode termio parameters */
    ioctl(0, TCGETS, &text_termio);

    graph_termio = text_termio;

    /* change termio parameters to allow our own I/O processing */
    graph_termio.c_iflag &= ~(BRKINT | PARMRK | INPCK | IUCLC | IXON | IXOFF);
    graph_termio.c_iflag |= (IGNBRK | IGNPAR);

    graph_termio.c_oflag &= ~(ONOCR);

    graph_termio.c_lflag &= ~(ICANON | ECHO | ECHOE | ECHOK | ECHONL | NOFLSH);
    graph_termio.c_lflag |= (ISIG);	/* enable interrupt */

    graph_termio.c_cc[VMIN] = 1;
    graph_termio.c_cc[VTIME] = 0;
    graph_termio.c_cc[VSUSP] = 0;	/* disable suspend */

/*    disable_interrupt();	/* Is reenabled later by set_texttermio */

/*    __svgalib_getchipset();		/* make sure a chipset has been selected */
/*    chipset_unlock();  */

    /* disable text output to console */
    ioctl(__svgalib_tty_fd, KDSETMODE, KD_GRAPHICS);

    __svgalib_takevtcontrol();	/* HH: Take control over VT */

    /* open /dev/mem */
    open_mem();

    /* mmap graphics memory */

    __vga_mmap();

    if ((long) GM < 0) {
	printf("svgalib: mmap error \n");
	exit(1);
    }
}
    
#define CRTR 0x3D4    
#define CRTC 0x18    
#define ATCR 0x3C0    
#define ATCC 0x15    
#define GRCR 0x3CE    
#define GRCC 0x9    
#define SECR 0x3C4    
#define SECC 0x5  
#define MISR 0x3C2
#define STAR 0x3DA    
delay()
{
   int i;
   for(i=0; i<10;i++);
}
setmode(byte *recs)
{
    int i;
    
    outw(0x11,CRTR); 
    for (i=0; i<CRTC; i++)
        outw(((*recs++)<<8)+i,(short)CRTR);
    inb(0x3DA);
    for (i=0; i<ATCC; i++)
    {
        delay();
        outb(i,(short)ATCR);
        delay();
        outb(*recs++,(short)ATCR);
    }
    inb(0x3DA);
    outb(0x20,ATCR);
    for (i=0; i<GRCC; i++)
        outw(((*recs++)<<8)+i,(short)GRCR);
    for (i=0; i<SECC; i++)

        outw(((*recs++)<<8)+i,(short)SECR);
    outb(*recs,(short)MISR);
}
    
static byte textmode[] = {
  0x5F,0x4F,0x50,0x82,0x55,0x81,0xBF,0x1F,	/* CR00-CR18 */
  0x00,0x4F,0x0E,0x0F,0x19,0xF0,0x1B,0x80,
  0x9C,0x00,0x8F,0x28,0x1F,0x96,0xB9,0xA3,
  0x00,0x01,0x02,0x03,0x04,0x05,0x06,0x07,	/* AR00-AR15 */
  0x08,0x09,0x0A,0x0B,0x0C,0x0D,0x0E,0x0F,
  0x0C,0x00,0x0F,0x08,0x00,
  0x00,0x00,0x00,0x00,0x00,0x10,0x0E,0x00,	/* GR00-GR08 */
  0xFF,
  0x03,0x00,0x03,0x00,0x02,			/* SR00-SR05 */
  0x67						/* MISC_OUT  */
};

static unsigned char mode640x480x16[] = {
  0x5F,0x4F,0x50,0x82,0x54,0x80,0x0B,0x3E,	/* CR00-CR18 */
  0x00,0x40,0x00,0x00,0x00,0x00,0x00,0x00,
  0xEA,0x8C,0xDF,0x28,0x00,0xE7,0x04,0xE3,
  0x00,0x01,0x02,0x03,0x04,0x05,0x06,0x07,	/* AR00-AR15 */
  0x08,0x09,0x0A,0x0B,0x0C,0x0D,0x0E,0x0F,
  0x01,0x00,0x0F,0x00,0x00,
  0x0F,0x0F,0x00,0x00,0x00,0x00,0x05,0x0F,	/* GR00-GR08 */
  0xFF,
  0x03,0x01,0x0F,0x00,0x06,			/* SR00-SR05 */
  0xE3,						/* MISC_OUT  */
};
#define FONTSIZE 32*256
byte *fontbuff;
fontread()
{
   byte *fb;
   int i;
   fontbuff=malloc(FONTSIZE*2);
   fb=fontbuff;
   outw(0x0005,GRCR);
   outw(0x0204,GRCR);
   for ( i=0;i<FONTSIZE; i++)
      *fb++ = GM[i];
   outw(0x0304,GRCR);
   for ( i=0;i<FONTSIZE; i++)
      *fb++ = GM[i];
   outw(0x0004,GRCR);      
}

fontwrite()
{
   byte *fb;
   int i;
   fb=fontbuff;
   outw(0x0001,GRCR);
   outw(0x0402,SECR);
   for ( i=0;i<FONTSIZE; i++)
      GM[i]=*fb++;
   outw(0x0802,SECR);
   for ( i=0;i<FONTSIZE; i++)
      GM[i]=*fb++;
   outw(0x0004,GRCR);      
}
/*
barwrite(int line,int value)
{
   int i;
   for(i=80*8*line; i<80*8*(line+1);i++)
      GM[i]= 1;
}*/
static byte bitmask[]={0xFF,0x7F,0x3F,0x1F,0xF,7,3,1,0};
/*static byte bitmaskr[]={0x0,0x80,0xC0,0xE0,0xF0,0xF8,0xFC,0xFE,0xFF};*/
Wbitmap0(int x,int y,int dx,int dy,int color)
{
   byte *ps,tmp;
   int col,ln;
   word lm,rm;
   
   dx+=tmp=x&7; 
   lm = (bitmask[tmp]<<8)+0x8;
   rm = (~bitmask[dx&7]<<8)+0x8;
   x >>= 3;
   dx >>= 3;
   ps = GM+x+y*width;
   outw(0x00+(color<<8),GRCR);
   if (dx)
      for(ln=0; ln<dy; ln++)
      {
         outw(lm,GRCR);
         *ps = *ps;
         outw(0xFF08,GRCR);
/*      for(col=0; col<dx-1; col++,ps++) *ps=*ps;*/
         asm(
         "    push %%ecx                 \n"
         "    push %%edi                 \n"
         "    dec %%ecx                  \n"
         "    jecxz bar_loop2            \n"
         "bar_loop:                      \n" 
         "    inc  %%edi                 \n"
         "    movb (%%edi),%%al          \n"
         "    movb %%al,(%%edi)          \n"
         "    loop bar_loop              \n"
         "bar_loop2:                     \n"
         ::
         "D"(ps),
         "c"(dx)
         );
         outw(rm,GRCR);
         asm(
         "    inc  %%edi                 \n"
         "    movb (%%edi),%%al          \n"
         "    movb %%al,(%%edi)          \n"
         "    pop  %%edi                 \n"
         "    pop  %%ecx                 \n"
         ::);
/*      ps[dx] = ps[dx];*/
         ps += width;   
      }
   else
   {
      lm &= rm;
      outw(lm,GRCR);
      for(ln=0; ln<dy; ln++)
      {
         (*ps)++;
         ps+=width;
      }
   }
   outw(0xFF08,GRCR);
   outw(0x0F00,GRCR);
}

Wbitmap1(int x,int y,int dx,int dy,
        int bx, int by, int dbx, int dby,
        int color,byte *bitmap)
{
   byte *ps,*bm;
   int col,ln,px;
   byte lm,rm,shift;
   
//   px=((x+dx)>>3)-(x>>3);   //
   dx+=x&7;
   if ((x&7)<(bx&7))                    /*Underflow */
      dx+=8;
   x-=bx&7;
   shift=x&7;  
   lm = bitmask[bx&7];
   rm = ~bitmask[(dx&7)]; 
//   dx = px;  //
   x  >>= 3;
   dx >>= 3;
   if (dbx&7)
   {
      dbx >>=3;
      dbx++;
   }   
   else
      dbx >>=3;
   bx >>=3;
   ps = GM+x+y*width;
   bm = bitmap+bx+by*dbx;
   outw(0x00+(color<<8),GRCR);
   if (dx)
      for(ln=0; ln<dy; ln++)
      {
/*        outw(0x08+(*bitmap++<<8),GRCR);*/
/*      for(col=0; col<dx; col++,ps++)*/
	 asm(
	    "    push %%ecx                 \n"
	    "    push %%ebx                 \n"
	    "    push %%esi                 \n"
	    "    push %%edi                 \n"
	    "    xorl %%eax,%%eax           \n"
	    "    movb %0,%%ah               \n"
	    "bit1_loop:                     \n"
	    "    push %%ecx                 \n"
	    "    andb (%%esi),%%ah          \n"
	    "    inc  %%esi                 \n"
	    "    movb %%bl,%%cl             \n"
	    "    shrl %%cl,%%eax            \n"
            "    orb  %%bh,%%ah             \n"    //
//            "    cmp  $0,%%al               \n"
//            "    je   nowrite               \n"
	    "    movb %%al,%%bh             \n"    //
	    "    movb  $8,%%al              \n"
	    "    outw %%ax,%%dx             \n"
	    "    incb (%%edi)               \n"
//	    "    movb %%cl,%%al             \n"    
            "nowrite:                       \n"
//	    "    roll $16,%%eax             \n"
	    "    xorl %%eax,%%eax           \n"
	    "    movb $255,%%ah             \n"    
	    "    inc  %%edi                 \n"
	    "    pop  %%ecx                 \n"
	    "    loopne bit1_loop           \n"
//	    "    xorb %%ah,%%ah             \n"
//	    "    rorl $8,%%eax              \n"
	    "    movb (%%esi),%%ah          \n"    //x
	    "    movb %%bl,%%cl             \n"    //x
	    "    shrl %%cl,%%eax            \n"    //x
            "    orb  %%bh,%%ah             \n"    //x
	    "    andb %1,%%ah               \n"
	    "    movb $8,%%al               \n"
	    "    outw %%ax,%%dx             \n"
	    "    incb (%%edi)               \n"
//  	    "    movw $8008,%%ax            \n"   //x
//	    "    outw %%ax,%%dx             \n"   //x
//	    "    incb (%%edi)               \n"   //x
	    "    pop  %%edi                 \n"
	    "    pop  %%esi                 \n"
	    "    pop  %%ebx                 \n"
	    "    pop  %%ecx                 \n"
	    ::
	    "m"(lm),
	    "m"(rm),
	    "D"(ps),
	    "S"(bm),
	    "c"(dx),
	    "d"((word)(GRCR)),
	    "b"(shift)
	    );
	 ps += width;
	 bm+=dbx;
      }
   else
   {
      
      for(ln=0; ln<dy; ln++)
      {
/*        outw(0x08+(*bitmap++<<8),GRCR);*/
/*      for(col=0; col<dx; col++,ps++)*/
	 asm(
	    "    push %%ecx                 \n"
	    "    push %%esi                 \n"
	    "    push %%edi                 \n"
	    "    xorl %%eax,%%eax           \n"
	    "    movb %0,%%ah               \n"
	    "    andb (%%esi),%%ah          \n"
	    "    rorl %%cl,%%eax            \n"
	    "    andb %1,%%ah               \n"
	    "    movb $8,%%al               \n"
	    "    outw %%ax,%%dx             \n"
	    "    incb (%%edi)               \n"
	    "    pop  %%edi                 \n"
	    "    pop  %%esi                 \n"
	    "    pop  %%ecx                 \n"
	    ::
	    "m"(lm),
	    "m"(rm),
	    "D"(ps),
	    "S"(bm),
	    "d"((word)(GRCR)),
	    "c"(shift)
	    );
	 ps += width;
	 bm+=dbx;
      }
   }
   outw(0xFF08,GRCR);
   outw(0x0F00,GRCR);

}

Wbitmap4(int x,int y,int dx,int dy,
        int bx, int by, int dbx, int dby,
        int color,byte *bitmap)
{
   byte *ps,*ps1,*bm,*bm1;
   int col,ln,co,cnt,scnt,tr;
   dword mask,mask1;
   byte shift,sl;
   
   
   shift=x&7;  
   x >>= 3;
/*   dx >>= 1;*/
   tr=color<<8;
   sl = bx&1;
   if (dbx & 1)
      dbx++;
   dbx >>=1;
   bx >>=1;
   ps1=ps = GM+x+y*width;
   bm1=bm = bitmap+bx+by*dbx;
   mask1 = 0x8000>>shift;
   for(ln=0; ln<dy; ln++)
   {
      mask = mask1;
      co = *bm++;
      if (sl)
      {
         co <<=4;
         cnt = 1;
      }
      else
         cnt = 2;
      for(col=0; col<dx; col++)
      {
	 outw(mask + 0x8,GRCR);
         if (!cnt--)
         {
            co = *bm++;
            cnt =1;
         }
         co <<=4;
         if ((co&0xf00) != tr)
         {
            outw(co&0xF00,GRCR);
	    (*ps)++;
         }
	 mask >>= 1;
	 if (mask&0xff)
	 {
	    mask=0x8000;
            ps++;
	 }
      }
      bm=bm1+=dbx;
      ps = ps1+=width;
   }
   outw(0xFF08,GRCR);
   outw(0x0F00,GRCR);

}


readbm(int x,int y,int dx,int dy,byte *bitmap)
{
   byte *ps,*ps1,*bm,*bm1;
   int col,ln,co,cnt,scnt,plane;
   dword mask,mask1;
   byte shift;
   
   
   shift=x&7;  
   x >>= 3;
   ps1=ps = GM+x+y*width;
   bm1=bm = bitmap;
   mask1 = 0x80>>shift;
   for(ln=0; ln<dy; ln++)
   {
      mask = mask1;
      co =0;
      cnt=1;
      for(col=0; col<dx; col++)
      {
         for(plane=0x0304; plane >=0x0004; plane-=0x0100)
         {
            int dum;
            outw(plane,GRCR);
            for (dum=0; dum<0; dum++);
            co <<=1;
            if (*ps & mask)
               co |= 1;
/*            Wlog(59,"%d,%d,%d",co,*ps,mask);*/
         }
/*         co =0;*/
         if (!cnt--)
         {
            *bm++=co;
            cnt =1;
            co =0;
         }
         
	 mask >>= 1;
	 if (!mask)
	 {
	    mask=0x80;
	    ps++;
	 }
      }
      if (!cnt)
         *bm++=co;
      ps = ps1+=width;
   }
   outw(0x0004,GRCR);      
}

scrcpy(int x,int y, int dx, int dy, int desy)
{
   byte *sp,*dp;
   int x1,y1;
   x >>=3;
   dx >>=3;
   dx++;
   
   outw(0x0105,GRCR);
   sp = GM+y*width+x;
   dp = GM+desy*width+x;
   for (y1 = 0; y1 <dy; y1++)
   {
      for (x1 = 0; x1 <dx; x1++)
      {
         *dp++=*sp++;
      }
      sp+=width-dx;
      dp+=width-dx;
   }
   outw(0x0005,GRCR);
}

vgad_init(int argc)
{
   int i;
   printf("ENTER WISIO!");
   width=80;
   height=480;
   initialize();
   setioperm();
   setmode(mode640x480x16);
   fontread();
   if(argc>1)
   {
      int fh;
      fh = open("SVFONT",O_RDONLY);
      if (fh==-1)
      {
         fh = open("SVFONT",O_WRONLY+O_CREAT,0666);
         write(fh,fontbuff,FONTSIZE*2);
      }   
      else
      {
         read(fh,fontbuff,FONTSIZE*2);
         fontwrite();  
      }
      setmode(textmode);
      close(fh);
      exit(0);
   } 
   outw(0x0000,GRCR);
   for (i=0; i <width*height; i++)
   {
        GM[i] =  0;
   }
   outw(0x0F00,GRCR);
 
}

static void vgad_exit(void)
{
   printf("START VGAD_EXIT");
   setmode(mode640x480x16);
   fontwrite();
   setmode(textmode);
   __vga_atexit();
   printf("EXIT WISIO!");
}

setgmode(int mode, int flags)
{
     if (mode)
        vgad_init(0);
     else
        vgad_exit();
}

drawbm(int bits,int color,
        int x,int y,int dx,int dy,
        int bx, int by, int dbx, int dby,
        byte *bitmap)
{
    int x1,y1,dx1,dy1,bx1,sy;
    
    if (bits==WBTP_CO1)
       return 
       Wbitmap0(x,y,dx,dy,color);
    sy=0;   
    for (y1=y; y1<y+dy; y1+=dy1)
    {
       dy1=dby-by;
       if (y1+dy1>y+dy)
	  dy1=y+dy-y1;
       if (dy1<0)
          return;
       bx1=bx;
       for (x1=x; x1<x+dx; x1+=dx1)
       {
	  dx1=dbx-bx1;
	  if (x1+dx1>x+dx)
	     dx1=x+dx-x1;
          if (dx1<0 )
             return;
          if (color==-1 && sy && x1>x+8 && x1+dx1+8 < x+dx)
             scrcpy(x1,sy,dx1,dy1,y1);
          else
	     switch(bits)
	     {
		case WBTP_CO2:
		   Wbitmap1(x1,y1,dx1,dy1,bx,by,dbx,dby,color,bitmap);
		   break;
		case WBTP_CO4:
		   break;
		case WBTP_CO16:
		   Wbitmap4(x1,y1,dx1,dy1,bx1,by,dbx,dby,color,bitmap);
		   break;
		case WBTP_CO256:
		case WBTP_CO64k:
		case WBTP_CO32M:
		   break;
	     }
   	  bx1=0;
       }
       if(by==0&&sy==0)
          sy=y1;
       by=0;
    }
}
/*readbm()
{
}
/* EXPORT FUNCTIONS */
extern int drawtria();
extern int D_Wbitmaps0();
int GRD_INIT(int flags, void *func[])
{
   if (flags & 1)
      vgad_init(2);
   func[ GRD_MODE ] = setgmode;
   func[ GRD_DRAW ] = drawbm;
   func[ GRD_READ ] = readbm;
   func[ GRD_TRIA ] = drawtria;
   func[ GRD_SHAPE ] = D_Wbitmaps0;
}
